
/*
   Programme pc_rm3 : régleur de montre - chronocomparateur
   Auteur : P. Chour - 2019-03-01 - V1.1
   Matériel : carte Arduino Mega, afficheur Touchscreen 3,5 pouces, 2 led, signal d'entrée
   V1.0 2019-03-01
   V1.1 2019-03-10 Différences par rapport à la version 1.0 :
    - suppression affichage si valeurs inchangées
    - amélioration algorithme de mesure

   V2.0 2019-03-24 : modifications importantes suite à changement de matériel (suppression du NE555. Entrée directe de l'amplificateur sur l'Arduino. Voir schéma (www.pascalchour.fr)
   - désormais, l'occultation est complètement réalisée par logiciel. En particulier, suppression du NE555 remplacé par une constante de temps minmum de 50ms en soft
     pour éliminer le train d'ondes dû aux battements.
   - introduction d'un automate de synchronisation de l'appareil pour se positionner sur le dégagement de l'échappement.
   - introduction d'un indicateur de synchronisation (vert = ok, rouge = pas synchronisé)
   - Les différences entre le Tic et le Tac sont moyennées depuis le début de la mesure alors que sur PC-RM1, il l'était sur les MaxMesures dernières mesures. La justification est que du fait
     de l'affichage graphique, on voit immédiatement une modification du réglage donc plus besoin d'avoir une valeur "instantannée" de la différence entre le Tic et le Tac. Par contre, 
     on est "content" d'avoir une moyenne en valeur qui se stabilise dans le temps !
   - les traitements sur les écarts et des affichages ont été simplifiés afin de diminuer les durées de calcul. 

  V2.1 2019-04-29 : CHANGEMENT MAJEUR
  - Relié à une carte Arduino pro micro qui fait les mesures et les transmet à la carte Mega. Du coup, la carte Mega 
    ne traite plus le signal d'entrée ni les LED mais reçoit les communications sur le port série. Tout l'algorithme de mesure
    est désormais traité par une carte Arduino pro micro qui transmet les valeurs à la carte Mega.

*/


#include <limits.h>
#include <EEPROM.h>
#include "AffichageTS35.h"  

//#define DEBUG_SERIAL
//#define DEBUG_ES
//#define DEBUG_AUTOMATE
//#define DEBUG_AFF
//#define SERIAL_MANUEL


//
// Constantes, define et variables pour la mesure.
//
#define MaxBattements   7             // Nombre de valeurs dans Battements[] (et idem pour Periode, demi periode et Aberrant)
#define MaxOccultations  5            // Nombre max de valeurs pour l'occultation
#define DureeSansSignal 50            // Pour la synchro. Délai durant lequel on ne veut pas de signal (en ms).
#define MaxInactivite 2000            // Temps au bout duquel on demande à la carte de mesure si elle est toujours là (en ms).
//Entêtes de protocole
#define EFinData 0x7F                 // fin d'une trame de donnée
#define ECartePresente 0x3F           // Interrogation et réponse attendue pour vérifier que la carte est vivante
#define ED1 0x1F                      // Envoie d'une durée d'occultation basse ou réception de la durée d'un TIC
#define ED2 0x2F                      // Envoie d'une durée occultation haute ou réception de la durée d'un TAC
#define EAckOccultation 0x4F          // Acquittement Occultation
#define ERien 0x5F                    // Trame non reçue ou incorrecte


long     EcartDureeTicTac;            // Conservation de la moyenne court terme de la différence entre un TicTac et la période attendue
long     NbEcartDureeTicTac;
uint64_t EcartDureeTicTacAncien;

int64_t  EcartDureeTicTacLong;
int64_t  NbEcartDureeTicTacLong;
uint64_t EcartDureeTicTacLongAncien;

uint64_t DureeLongTerme;     // Pour mesure précise des périodes. En µS.
uint64_t NbDureeLongTerme;   // Nombre de mesures DureeLongTerme
uint64_t AncienAR;

uint64_t EcartTicTac;        // Pour moyenne de la différence de l'écart entre Tic et Tac
uint64_t NbEcartTicTac;      // pour calcul de la moyenne : nombre d'EcartTicTac 
uint64_t EcartTicTacAncien;

unsigned long TimerInactivite; // Pour mesurer le temps d'inactivité de la carte de mesure et vérifier si elle répond



// Les constantes qui suivent dépendent de Battements. Elles sont précalculées pour éviter de perdre du temps lors des traitements.
// Si on souhaite ajouter un nouveau battement, on l'ajoute dans le tableau Battement à la position i
// On calcule la période correspondante que l'on ajoute dans le tableau Periode à la position i
// On calcule la demi période que l'on ajoute dans le tableau DemiPeriode à la position i

// On met à jour la valeur MaxBattements qui correspond aux nombres de battements que l'on a défini.
const int64_t Battements[MaxBattements] = {18000, 19800, 21600, 25200, 28800, 36000, 3600};             // en battements par heures
const int64_t Periode[MaxBattements] = {400000, 363636, 333333, 285714, 250000, 200000, 1000000};       // periode en µS
const int64_t DemiPeriode[MaxBattements] = {200000, 181818, 166666, 142857, 125000, 100000, 500000};    // demi période en µS

// Mesures des temps (temps courant, durée d'un tic ou d'un tac courant)
unsigned long DiffTicTac;                // Durée d'un Tic ou d'un Tac. 
unsigned long DiffTicTacPrec;            // Valeur précédente de DiffTicTac

// Tableau des boutons de l'affichage de l'écran de mesure
const TBouton BoutonAffichageMesure1[4] = {{289,10,LargeurCol,"Reprise",ROUGE,BLANC},
              {289,20+LargeurCol,LargeurCol,"Reinit.",BLANC,CouleurFond,ROUGE,BLANC},
              {289,30+2*LargeurCol,LargeurCol,"Battement",BLANC,CouleurFond,ROUGE,BLANC},
              {-1,-1,0,"",0,0,0,0}};
const TBouton BoutonAffichageMesure2[4] = {{289,10,LargeurCol,"Pause",BLANC,CouleurFond,ROUGE,BLANC},
              {289,20+LargeurCol,LargeurCol,"",CouleurFond,CouleurFond,ROUGE,BLANC},
              {289,30+2*LargeurCol,LargeurCol,"",CouleurFond,CouleurFond,ROUGE,BLANC},
              {-1,-1,0,"",0,0,0,0}};
              
const TBouton BoutonAffichageBattement[MaxBattements+1] = {{60,40,LargeurCol,"18000",BLANC,CouleurFond,ROUGE,BLANC},
              {60,50+LargeurCol,LargeurCol,"19800",BLANC,CouleurFond,ROUGE,BLANC},
              {60,60+2*LargeurCol,LargeurCol,"21600",BLANC,CouleurFond,ROUGE,BLANC},
              {100,40,LargeurCol,"25200",BLANC,CouleurFond,ROUGE,BLANC},
              {100,50+LargeurCol,LargeurCol,"28800",BLANC,CouleurFond,ROUGE,BLANC},
              {100,60+2*LargeurCol,LargeurCol,"36000",BLANC,CouleurFond,ROUGE,BLANC},
              {140,40,LargeurCol,"3600",BLANC,CouleurFond,ROUGE,BLANC},
              {-1,-1,0,"",0,0,0,0}};

const TBouton BoutonOccultation[MaxOccultations+1] = {{260,30,PetitBouton,"50ms",BLANC,CouleurFond,ROUGE,BLANC},
              {260,40+PetitBouton,PetitBouton,"1/4",BLANC,CouleurFond,ROUGE,BLANC},
              {260,50+2*PetitBouton,PetitBouton,"1/2",BLANC,CouleurFond,ROUGE,BLANC},
              {260,60+3*PetitBouton,PetitBouton,"3/4",BLANC,CouleurFond,ROUGE,BLANC},
              {260,70+4*PetitBouton,PetitBouton,"9/10",BLANC,CouleurFond,ROUGE,BLANC},
              {-1,-1,0,"",0,0,0,0}};
              
const TBouton BoutonCommunication[2] = {{289,400,PetitBouton,"C",VERT,CouleurFond,ROUGE,BLANC},
              {-1,-1,0,"",0,0,0,0}};   

const char Depassement[4] = "***\0";

// Etats de l'automate
//
enum TEtat {E_selection,E_mesure,E_selection_trans,E_mesure_trans,E_pause,E_pause_in_trans,E_mesure_redemarre_trans,E_selection_change_trans};


/*
   Initialisation du programme
*/
void setup()
{

#ifdef DEBUG_SERIAL || DEBUG_ES || DEBUG_AUTOMATE
  Serial.begin(9600);  
#endif
  USART_Init_76800();               // init E/S avec la carte pro micro qui réalise les mesures.
  AfficheInit();                    // Initialisation de l'écran graphique
  AfficheEcranAccueil();            // Affichage de l'écran d'accueil
  delay(1000);
}


/*
 * Change l'état du bouton selon que la carte est dispo ou pas
 * Force à vrai => force l'affichage
 * Force à faux => n'affiche que si l'état change
 */
void MAJBoutonComm(bool *CarteMesureVivante, bool Force) {
bool B;  
  if (DureeVrai(TimerInactivite, millis()) > MaxInactivite) {
    USART_Transmit_Char(ECartePresente);
    TimerInactivite = millis();
    B = USART_Available();
  } else {
    B = true;
  }
  if (B) {
    if (!(*CarteMesureVivante) || Force) {AfficheUnBouton(&BoutonCommunication[0], false);} // On indique s'il y a une communication avec le pro-micro
    *CarteMesureVivante = true;
  } else {
    if (*CarteMesureVivante || Force) {AfficheUnBouton(&BoutonCommunication[0], true);} // On indique s'il y a une communication avec le pro-micro
    *CarteMesureVivante = false;
  } 
}

/*
 * Réception par test d'état ou via les routines standard selon
 * SERIAL_MANUEL. Pour la carte de mesure, il est impératif que les E/S
 * se fassent par test d'état => SERIAL_MANUEL doit être défini.
 * Pour la carte d'affichage, on peut utiliser l'un ou l'autre mode.
 * Utilisation de Serial1 (TX1 et RX1).
 */

 /*
  * Réception d'un caractère. La routine est boquante.
  */
 unsigned int USART_Receive()
{
#ifdef DEBUG_ES
  unsigned int C;
  #ifdef SERIAL_MANUEL  
    while (!(UCSR1A & (1<<RXC1)));
    C = UDR1;
  #else
    while (Serial1.available() == 0);
    C = Serial1.read();
  #endif
    Serial.print("-> ");
    Serial.println(C,HEX);
    return C;  
#else
  #ifdef SERIAL_MANUEL  
    while (!(UCSR1A & (1<<RXC1)));
    return UDR1;
  #else
    while (Serial1.available() == 0);
    return Serial1.read();
  #endif
#endif
}

/*
 * Indique si un caractère a été reçu
 */
bool USART_Available() {
bool B;  
#ifdef SERIAL_MANUEL  
  B = (UCSR1A & (1<<RXC1));
#else
  B = (Serial1.available() > 0);
#endif  
  if (B) {TimerInactivite = millis();}
  return B;
}

/*
 * Attend jusqu'à ce que la transmission soit terminée (TXC1 = 1)
 */
void USART_Flush_Tx()
{
#ifdef SERIAL_MANUEL  
  while (!(UCSR1A & (1<<TXC1)));
  delay(5); // On laisse le temps à la ligne de se stabiliser
#else
  Serial1.flush();
#endif  
}

/*
 * Emission d'un caractère
 */
void USART_Transmit_Char(unsigned int data)
{
#ifdef SERIAL_MANUEL   
  while (!( UCSR1A & (1<<UDRE1))); // Tant que le buffer n'est pas vide, on boucle
  UDR1 = data;  // transmission
#else
  USART_Flush_Tx();  
  Serial1.write((byte)(data & 0xFF));
#endif  
#ifdef DEBUG_ES
  Serial.print("<- ");
  Serial.print(((data >> 4) & 0x0F), HEX);
  Serial.println((data & 0x0F), HEX);
#endif
}

void USART_Init_76800()
{
#ifdef SERIAL_MANUEL  
  UCSR1B = 0;                         // Désactive UART en émission/réception
  //UCSR1D = 0;                       // pas de CTS ni de RTS
  UCSR1C = (1<<UCSZ11)|(1<<UCSZ10);   // 8 bits 1 stop pas de parité
  UCSR1A = (1<<U2X1);                 // double vitesse. Permet d'avoir une meilleure précision en bps
  UBRR1H = 0;                         // poids fort du diviseur
  UBRR1L = 25;                        // = (16000000 / 4 / 76800 - 1 ) / 2
  UCSR1B = (1<<RXEN1)|(1<<TXEN1);     // active l'émission et la réception
#else  
  Serial1.begin(76800,SERIAL_8N1);     // init E/S avec la carte pro micro qui réalise les mesures.
#endif  
}

/*
 * Vide le buffer de réception. Utilisé dans les états ou on n'attend rien
 */
void VideReception() {
unsigned int i;
  while (USART_Available()) {i = USART_Receive();} // Vide le buffer de réception
}
/*
 * Attend une trame data. Rend vrai si reçue. Dans ce cas, R1 et R2 contiennent les valeurs reçues.
 */

bool AttenteTrameData(unsigned long *R1, unsigned long *R2) {
unsigned long Timer;
unsigned int Recu;
unsigned int Recu1;
unsigned int Recu2;
unsigned long V1;
unsigned long RTemp;
bool BED1;
bool BED2;
  BED1 = false;
  BED2 = false;
  Timer = millis();  
  Recu = ERien;
  while ((DureeVrai(Timer,millis()) < 500) && (Recu != ED1)) {
    if (USART_Available()) {
      Recu = USART_Receive();
    }
  }
  if (Recu == ED1) {
    RTemp = 0;
    V1 = 1;    
    Timer = millis();  
    while ((DureeVrai(Timer,millis()) < 500)) {
      if (USART_Available()) {
        Recu = USART_Receive();
        switch (Recu) {
          case ECartePresente: // On filtre certains éléments de protocoles
          case EAckOccultation:
          case ED1:
            return false;
          break;
          case EFinData:
            if (BED2) {
              *R2 = RTemp;
              return true;
            } else {
              return false;
            }
          break;
          case ED2:
            if (BED2) {
              return false; 
            } else {
              *R1 = RTemp;
              BED2 = true;
              RTemp = 0;
              V1 = 1;
            }
          break;
          default:
            Recu1 = Recu & 0x0F;
            Recu2 = (Recu >> 4) & 0x0F;
            if ((Recu1 <= 9) && (Recu2 <= 9)) {
              RTemp = RTemp + (Recu1+Recu2*10)*V1;
              V1 = V1*100;
            } else {
              return false;
            }
          break;    
        }
      }          
    }
  }
  return false;
}  


/*
 * Emission Fenetre occultation
 * En entrée : fenetre basse et haute
 * La transmission se fait poids faible en tête pour chaque octet et poids faible en tête pour la trame
 * 150 => transmission 50 puis transmission 01
 * 39876 => transmission 76 puis transmission 98 puis transmission 03
 */
 bool EmissionFenetreOccultation(unsigned long T1,unsigned long T2) {
 unsigned int i;
 unsigned long Timer;
#ifdef DEBUG_ES
  Serial.print("Emet fenêtre : ");
  Serial.print(T1);
  Serial.print(" - ");
  Serial.println(T2);
#endif
  USART_Transmit_Char(ED1);
  do {
    i = (T1-((T1 / 10)*10)); // Attention aux optimisation !
    T1 = T1/10;
    i = ((T1-((T1 / 10)*10)) << 4) + i; // Attention aux optimisation !
    T1 = T1/10;
    USART_Transmit_Char(i);
  } while (T1 != 0);
  USART_Transmit_Char(ED2);
  do {
    i = (T2-((T2 / 10)*10)); // Attention aux optimisation !
    T2 = T2/10;
    i = ((T2-((T2 / 10)*10)) << 4) + i; // Attention aux optimisation !
    T2 = T2/10;
    USART_Transmit_Char(i);
  } while (T2 != 0);
  VideReception();
  USART_Transmit_Char(EFinData);

  Timer = millis();  
  while (DureeVrai(Timer,millis()) < 500) {
    if (USART_Available()) {
      if (USART_Receive() == EAckOccultation) {
        return true;
      }
    }
  }
  return false;
 }

/*
 * Initialise la fenêtre d'occultation en fonction de l'index de la fenêtre choisie et de l'index du battement courant
 * En entrée : index de la fenêtre d'occulation. Pas de contrôle !
 *             index du battement choisi. Pas de contrôle
 * En sortie : Fenêtre basse d'occultation et fenêtre haute d'occultation initialisés (en ms)
*/
void InitFenetre(int Ioccult, int IndexBattements, unsigned long *FenetreBasseOccultation, unsigned long *FenetreHauteOccultation)
{
  switch (Ioccult) {
    case 1 : *FenetreBasseOccultation = DemiPeriode[IndexBattements]/4000; *FenetreHauteOccultation = (Periode[IndexBattements]/1000)-*FenetreBasseOccultation; break;
    case 2 : *FenetreBasseOccultation = DemiPeriode[IndexBattements]/2000; *FenetreHauteOccultation = (Periode[IndexBattements]/1000)-*FenetreBasseOccultation; break;
    case 3 : *FenetreBasseOccultation = (DemiPeriode[IndexBattements]*3)/4000; *FenetreHauteOccultation = (Periode[IndexBattements]/1000)-*FenetreBasseOccultation; break;
    case 4 : *FenetreBasseOccultation = (DemiPeriode[IndexBattements]*9/10000); *FenetreHauteOccultation = (Periode[IndexBattements]/1000)-*FenetreBasseOccultation; break;
   default : *FenetreBasseOccultation = 50; *FenetreHauteOccultation = (Periode[IndexBattements]/1000)-50; break; 
  }      
}  


//
// La fonction ci-dessous mesure le temps en prenant en compte un éventuel bouclage de la fonction millis ou micro
// On lui donne la valeur de départ de la valeur à mesurer et la valeur courante de millis ou micro. Elle rend la durée écoulée en millisecondes ou microseconde.
// DureeVrai = courante-dernière
//
unsigned long DureeVrai(unsigned long derniere, unsigned long courante) {
  if (courante < derniere) {  // on est passé par zéro
    return (0xffffffffu - derniere) + courante;
  } else {
    return courante - derniere;
  }
}

//
// Formatte une valeur pour affichage en microsecondes, millisecondes, secondes ou minutes dans un tableau de char terminé par null
// V1 = valeur en microsecondes
// L = longueur max affichage. Le formatage implique au moins 2 caractères + éventuellement un signe => jusqu'à 3 caractères sont pris. L Max = 19
// Signe = Signe à afficher. Espace => pas de signe. On gagne un caractère sur l'affichage.
// Rend "***" si dépassement capacité ou affichage impossible
// Affichage => unité sur deux caractères + signe sur 1 (+ ou -) ou (pas de signe) caractère + 1 chiffre => L >= 3 si pas de signe et L >= 4 si signe
// Pas de vérification de débordement !!! L'appelant doit faire attention que L et le buffer soient de la bonne taille.
//
void FormatTemps(uint64_t V1, int L, char Signe, char buf[]) {
uint64_t VTempPF;
uint64_t VTemppf;
char bufPF[20];
int iPF, k, i, j;
    if (L > 20) {L=19;}  // Une précaution !
    buf[L] = '\0';
    for (i=0; i<L;i++){bufPF[i]=' ';}
    
    VTempPF = V1/60000000;
    if (VTempPF > 0) {
      buf[L-1] = 'n';
      buf[L-2] = 'm';
      VTemppf = (V1-(VTempPF*60000000))/1000000;
    } else {
      VTempPF = V1/1000000;
      if (VTempPF > 0) {
        VTemppf = V1-(VTempPF*1000000)/1000;
        buf[L-1] = 's';
        buf[L-2] = ' ';
      } else {
        VTempPF = V1/1000;
        if (VTempPF > 0) {
          VTemppf = (V1-VTempPF*1000);
          buf[L-1] = 's';
          buf[L-2] = 'm';
        } else {
          VTemppf = V1;
          buf[L-1] = 's';
          buf[L-2] = 'u';
        }
      }
    }
    if (((VTempPF > 10) && ((L-2) <= 1)) || ((VTempPF > 100) && ((L-2) <= 2)) || (VTempPF > 1000)) {
        memcpy(buf,Depassement,sizeof(Depassement));
        return;
    }    
    iPF = 0;
    while (VTemppf > 0) {
      bufPF[iPF] = (char)(48+VTemppf-(VTemppf/10)*10); // attention aux optimisations du compilo !
      VTemppf = VTemppf/10;
      iPF = iPF+1;
    }
    if (iPF == 0) {
      bufPF[0] = '0';
      iPF = 1;
    }

    bufPF[iPF] = '.';
    iPF = iPF+1;
    if (VTempPF == 0) {
      bufPF[iPF] = '0';
      iPF = iPF+1;
    } else {
      while (VTempPF > 0) {
        bufPF[iPF] = (char)(48+VTempPF-(VTempPF/10)*10); // attention aux optimisations du compilo !
        VTempPF = VTempPF/10;
        iPF = iPF+1;
      }
    }
    if (Signe != ' ') {
      bufPF[iPF] = Signe;
      iPF = iPF+1;
    };   

    i = L-3;
    j = iPF-L+2;
    if (j < 0) {j = 0;}
    k = 0;
    for (k=j; k < iPF; k++) {
      buf[i] = bufPF[k];
      i = i-1;
    }
    while (i >= 0) {
      buf[i] = ' ';
      i = i-1;
    }
#if defined(DEBUG_AFF)      
    Serial.println(String((unsigned long)V1) + " / " + buf);
#endif    
}
//
// Affichage des mesures
// AAfficher² : 0 = Différence TicTac, 1=Différence période, 2= moyenne période, 3=avance retard par jour
//
void AffMesure(int IndexBattements, byte AAfficher)
{
  int i;
  uint64_t MoyenneTic;
  uint64_t MoyenneTac;
  long VTemp;
  uint64_t VTempU;
  int64_t  VTemp64;
  char buf[21];
  char Signe ;

  
// Affichage période. On n'affiche que si l'on a au moins deux mesures
    switch (AAfficher) {
      case 0 : // Moyenne des différences entre Tic et Tac    
            if (NbEcartTicTac > 0) {
              VTempU = EcartTicTac/NbEcartTicTac;
              if (VTempU != EcartTicTacAncien) {
                EcartTicTacAncien = VTempU; 
                FormatTemps(VTempU,8,' ',buf);
                AfficheBloc(5,CouleurFond,VERT,buf);
              }
            }
      break;      
      case 1 : // Moyenne de la différence entre la durée d'un Tic ou d'un Tac et la demi-période attendue
          if (NbEcartDureeTicTac > 0) {
            EcartDureeTicTacLong = EcartDureeTicTacLong+EcartDureeTicTac;
            NbEcartDureeTicTacLong = NbEcartDureeTicTacLong+NbEcartDureeTicTac;
            if (EcartDureeTicTac > 0) {
              VTempU = EcartDureeTicTac/NbEcartDureeTicTac;
              Signe = '+';
            } else {
              VTempU = (-EcartDureeTicTac)/NbEcartDureeTicTac;
              Signe = '-';
            }
            if (VTempU != EcartDureeTicTacAncien) {
              FormatTemps(VTempU,8,Signe,buf);
              AfficheBloc(6,CouleurFond,VERT,buf);              
              EcartDureeTicTacAncien = VTempU;
            }
            EcartDureeTicTac = 0;
            NbEcartDureeTicTac = 0;
          }
      break;
      case 2://Moyenne de la période
          if (NbEcartDureeTicTacLong > 0) {
            if (EcartDureeTicTacLong > 0) {
              VTempU = EcartDureeTicTacLong/NbEcartDureeTicTacLong;
              Signe = '+';
            } else {
              VTempU = (-EcartDureeTicTacLong)/NbEcartDureeTicTacLong;
              Signe = '-';
            }
            if (VTempU != EcartDureeTicTacLongAncien) {
              FormatTemps(VTempU,8,Signe,buf);
              EcartDureeTicTacLongAncien = VTempU;
            }            
            AfficheBloc(7,CouleurFond,VERT,buf);
          }
      break;
      case 3://Avance retard par jours
        VTempU = Periode[IndexBattements]*NbDureeLongTerme; 
        if (VTempU > DureeLongTerme) {                          // Si durée attendue est supérieure à la durée mesurée
          VTempU = (VTempU-DureeLongTerme)/NbDureeLongTerme;    // La montre avance
          Signe = '+';
        } else {
          VTempU = (DureeLongTerme-VTempU)/NbDureeLongTerme;    // Sinon, elle retarde    
          Signe = '-';
        }
        if (VTempU != AncienAR) {
          AncienAR = VTempU;              
          FormatTemps(VTempU*Battements[IndexBattements]*(uint64_t)12,8,Signe,buf);
          AfficheBloc(8,CouleurFond,VERT, buf);
        }
      break;
      default:;
    }
}


/*
 * APPLICATION
 * Utilise un automate d'Etats finis
 * L'automate comporte des Etats et des transitions menant à l'Etat
 * convention de nommage : E_<nom d'état> représente un Etat
 *                         E_<nom d'état>_Trans représente une transition pour aller dans l'état <nom état>
 * Le passage d'un état à l'autre est déclenché par des événements :                         
 *    - un appui sur l'écran pour sélectionner une option
 *    - la réception d'une trame...
 */

void loop()
{
  int i;
  bool CalculPossible = false; 
  byte AAfficher = 0;                     // pour affichage des mesures. Pour des raisons de temps d'affichage, on n'affiche qu'une mesure à la fois
  unsigned long AffTimer;                 // On ne fait un affichage des mesures que toutes les 1000ms (Arduino, c'est lent !)
  unsigned long CommTimerAck;             // Timer pour acquittement transmission fenêtre occultation
  bool CarteMesureVivante;                // Pour se souvenir si la carte de mesure répond
  unsigned long FenetreBasseOccultation;  // Durée basse à partir de laquelle on accepte une interruption
  unsigned long FenetreHauteOccultation;  // Durée haute au delà de laquelle on refuse accepte une interruption. On doit resynchroniser la mesure
  int Recu;
  unsigned long V1; 
  unsigned int Recu1;                     // pour calcul valeur DiffTicTac
  unsigned int Recu2;                     // pour calcul valeur DiffTicTac
  unsigned long DiffTicTacTemp;           // pour calcul valeur DiffTicTac
  long TempsTic;                          // pour affichages, temporaire
  long TempsTac;                          // pour affichages, temporaire
  TEtat EtatAutomate = E_selection_trans; // Etat courant de l'automate général

  int IndexBattements;                    // Valeur du battement courant dans le tableau Battements, Periode, demi-periode...
  int IndexOccultation;                   // Valeur d'occultation de la mesure par défaut (3/4)
  const unsigned int AdIndexBattements = 0;             // Adresse de la sauvegarde de IndexBattement en EEPROM
  const unsigned int AdIndexOccultation = AdIndexBattements+sizeof(IndexBattements);       // Adresse de la sauvegarde de IndexOccultation en EEPROM  
  #ifdef DEBUG_AUTOMATE
     TEtat OldEtatAutomate = E_selection_trans;
  #endif

 
  EEPROM.get(AdIndexBattements,IndexBattements); // Récupération des valeurs par défaut de IndexBattements, IndexOccultation et supr
  EEPROM.get(AdIndexOccultation,IndexOccultation);  
  if (IndexBattements >= MaxBattements) {IndexBattements = 0;} // Ca peut arriver si l'EEPROM n'est pas encore initialisée.
  if (IndexOccultation > MaxOccultations) {IndexOccultation = 0;}  
  InitFenetre(IndexOccultation,IndexBattements,&FenetreBasseOccultation,&FenetreHauteOccultation);
  AffTimer = millis();
  CarteMesureVivante = USART_Available();

    
  while (1) {
    
#ifdef DEBUG_AUTOMATE
   if (OldEtatAutomate != EtatAutomate) {
    OldEtatAutomate = EtatAutomate;
    switch(EtatAutomate) {
      case E_selection:
        Serial.println("1) E_selection");
      break;
      case E_mesure:
        Serial.println("1) E_mesure");
      break;
      case E_selection_trans:
        Serial.println("1) E_selection_trans");
      break;
      case E_mesure_trans:
        Serial.println("1) E_mesure_trans");
      break;
      case E_pause:
        Serial.println("1) E_pause");
      break;
      case E_pause_in_trans:
        Serial.println("1) E_pause_in_trans");
      break;
      case E_mesure_redemarre_trans:
        Serial.println("1) E_mesure_redemarre_trans");
      break;
      case E_selection_change_trans:
        Serial.println("1) E_selection_change_trans");
      break;
      default:
        Serial.println("1) Inconnu !");
      break;
    }
  }
#endif

    //-------------------------------------------------
    // Automate de l'application
    //-------------------------------------------------
    switch (EtatAutomate) {

      //------------------------------------------------------------------------------------
      // Passage en mode sélection, on efface l'écran et on passe en changement de battement
      //------------------------------------------------------------------------------------
      case E_selection_trans:
        AfficheEcranBattement(BoutonAffichageBattement, BoutonOccultation);
        AfficheUnBouton(&BoutonAffichageBattement[IndexBattements],true);
        AfficheUnBouton(&BoutonOccultation[IndexOccultation],true);
        MAJBoutonComm(&CarteMesureVivante,true);
        EtatAutomate = E_selection;
      break;

      //---------------------------------------------------------
      // Etat Sélection du battement et de la durée d'occultation
      //---------------------------------------------------------
      case E_selection:
        VideReception();
        MAJBoutonComm(&CarteMesureVivante,false);
        if (IsPressed()){
         
          i = NumBouton(BoutonAffichageBattement);
          if (i != -1) {
            AfficheUnBouton(&BoutonAffichageBattement[IndexBattements],false);            
            IndexBattements = i;
            ClignoteBouton(&BoutonAffichageBattement[IndexBattements]);
            EEPROM.put(AdIndexBattements,IndexBattements);   // Sauvegarde valeur par défaut en EEPROM. N'écrit que s'il y a eu changement
            InitFenetre(IndexOccultation,IndexBattements,&FenetreBasseOccultation,&FenetreHauteOccultation);
            // Transmission à la carte de mesure
            if (EmissionFenetreOccultation(FenetreBasseOccultation,FenetreHauteOccultation)) {
              EtatAutomate = E_mesure_trans;
            } else {
              AfficheUnBouton(&BoutonAffichageBattement[IndexBattements],true);
            }       
          } else {
            i = NumBouton(BoutonOccultation);
            if (i != -1) {
              AfficheUnBouton(&BoutonOccultation[IndexOccultation],false);
              IndexOccultation = i;
              AfficheUnBouton(&BoutonOccultation[IndexOccultation],true);
              EEPROM.put(AdIndexOccultation,IndexOccultation); // Sauvegarde valeur par défaut en EEPROM. N'écrit que s'il y a eu changement
              InitFenetre(IndexOccultation,IndexBattements,&FenetreBasseOccultation,&FenetreHauteOccultation);
            }
          }
        }
      break;      


      //---------------------------------------------------------------------------------------
      // Passage en mode mesure. On réinitialise ls mesures et on affiche les valeurs courantes
      //---------------------------------------------------------------------------------------
      case E_mesure_trans:
        AfficheEcranMesure(BoutonAffichageMesure2);
        MAJBoutonComm(&CarteMesureVivante,true);
        EcartDureeTicTac = 0;
        NbEcartDureeTicTac = 0;
        EcartTicTacAncien = 0;
        EcartDureeTicTacAncien = 0;
        EcartDureeTicTacLong = 0;
        NbEcartDureeTicTacLong = 0;
        EcartDureeTicTacLongAncien = 0;
        EcartTicTac = 0;
        NbEcartTicTac = 0;
        DureeLongTerme = 0;
        NbDureeLongTerme = 0; 
        AncienAR = 0;
        EtatAutomate = E_mesure;
      break;

      //-----------------
      // passage en pause
      //-----------------
      case E_pause_in_trans: // on n'était pas dans l'état E_pause
          AfficheBoutons(BoutonAffichageMesure1);  
          EtatAutomate = E_pause;
      break;

      //--------------------------------------------
      // Etat pause, Attente d'un événement (bouton)
      //--------------------------------------------
      case E_pause:
        VideReception();
        MAJBoutonComm(&CarteMesureVivante,false);
        if (IsPressed()){      
            switch (NumBouton(BoutonAffichageMesure1)) {
              case 0: // reprise des mesures
                  EtatAutomate = E_mesure_redemarre_trans;
              break;  
              case 1: // Réinitialisation des mesures
                  EtatAutomate = E_mesure_trans;
              break;
              case 2: // Sélection des battements
                  EtatAutomate = E_selection_trans;
              break;
            }
         }
      break;

      //-------------------------
      // Redémarrage de la mesure
      //-------------------------
      case E_mesure_redemarre_trans:
        AfficheBoutons(BoutonAffichageMesure2);
        EtatAutomate = E_mesure;
      break;

      //------------
      // Etat Mesure
      //------------
      case E_mesure:
        CalculPossible = AttenteTrameData(&DiffTicTacPrec, &DiffTicTac);      
        MAJBoutonComm(&CarteMesureVivante, false);
        if (CalculPossible) { // Calcul possible lorsqu'on a un DiffTicTac et un DifTicTacprec
          // Pour Calcul de la moyenne de la différence entre un TicTac et la période attendue
          EcartDureeTicTac = EcartDureeTicTac+((long)(DiffTicTac+DiffTicTacPrec)-Periode[IndexBattements]);
          TempsTic = (long)DiffTicTacPrec-DemiPeriode[IndexBattements];
          TempsTac = (long)DiffTicTac-DemiPeriode[IndexBattements];
          NbEcartDureeTicTac = NbEcartDureeTicTac+1;            

          // Pour calcul de la moyenne de la différence entre un Tic et un Tac
          if (DiffTicTacPrec > DiffTicTac) {EcartTicTac = EcartTicTac+DiffTicTacPrec-DiffTicTac;} else {EcartTicTac = EcartTicTac+DiffTicTac-DiffTicTacPrec;}
          NbEcartTicTac = NbEcartTicTac+1; 
          // Pour calcul de l'avance/retard
          if (DureeLongTerme < 0x0FFFFFFFFFFFFFFF) {
            DureeLongTerme = DureeLongTerme + DiffTicTac + DiffTicTacPrec;
            NbDureeLongTerme = NbDureeLongTerme+1;
          }
          AfficheTrace((TempsTic/100),(TempsTac/100));
          CalculPossible = false;
          if (DureeVrai(AffTimer, millis()) > 1000)  { // on affiche que toutes les 1000 ms 
            AffTimer = millis();
            AffMesure(IndexBattements, AAfficher);
            AAfficher = (AAfficher+1) %4;             // Et on affiche qu'une valeur à la fois car ça prend du temps ! La valeur à afficher change à chaque appel
        }
      }
      if (IsPressed()){
        if (NumBouton(BoutonAffichageMesure2) == 0) {
          EtatAutomate = E_pause_in_trans;
        }
      }      
      break;

//
// L'état default correspond à une erreur d'automate (pas d'état connu).
// En cas d'erreur, on retourne dans l'état E_Selection
//
      default:
        EtatAutomate = E_selection_trans;
      break;
    }



  }
}
