
/*
   Programme pc_rm3_capteur : régleur de montre - chronocomparateur
   Auteur : P. Chour - 2019-05-02 - V1.0
   Matériel : carte Arduino pro-micro, 2 LEDS, une entrée "signal", E/S asynchrone
   Fonctionne avec PC_rm3v2.1. et au delà Communication par liaison série

   Réalise la mesure temporelle des battements de la montre et transmet les valeurs à la carte Arduino Mega qui réalise les affichage sur
   écran graphique.
   - l'occultation est complètement réalisée par logiciel. (Suppression du NE555 de PC-RM1 remplacé par une constante de temps réglable)
     pour éliminer le train d'ondes dû aux battements.
   - introduction d'un automate de synchronisation de l'appareil pour se positionner sur le dégagement de l'échappement.
   - communication des mesures sur voie série.
   - réception des consignes sur voie série (consigne  = durée fenêtres d'occultation)

   Ce module a été introduit pour PC-RM3 à partir de la version 2.1 car la carte Arduino Mega n'est pas assez rapide pour traiter correctement les affichages et les mesures.
   En particulier, il semble que les affichages perturbent l'horloge et fait perdre quelques ms de précision.
   Dans ce module, il n'y a pas d'autres IT que la réception du signal de la montre et le timer. Les E/S se font par test d'état.


*/


#include <limits.h>
#include <avr/io.h>
#include <PinChangeInterrupt.h>
#include <PinChangeInterruptBoards.h>
#include <PinChangeInterruptPins.h>
#include <PinChangeInterruptSettings.h>


//#define DEBUG_SERIAL
#define SERIAL_MANUEL

//
// Constantes, define et variables pour la mesure.
//
#define DureeSansSignal 50            // Pour la synchro. Délai durant lequel on ne veut pas de signal (en ms).
//Entêtes de protocole
#define EFinData 0x7F                 // fin d'une trame de donnée
#define ECartePresente 0x3F           // Interrogation et réponse attendue pour vérifier que la carte est vivante
#define ED1 0x1F                      // Envoie d'une durée d'occultation basse ou réception de la durée d'un TIC
#define ED2 0x2F                      // Envoie d'une durée occultation haute ou réception de la durée d'un TAC
#define EAckOccultation 0x4F          // Acquittement Occultation


// Mesures des temps (temps courant, durée d'un tic ou d'un tac courant)
unsigned long TempsTicTacPrec;            // Valeur précédente de la valeur courante du timer
volatile unsigned long TempsTicTac;       // Valeur courante du Timer. Mis à jour par IT
unsigned long DiffTicTac;                 // Durée d'un Tic ou d'un Tac.
unsigned long DiffTicTacPrec;             // Valeur précédente de DiffTicTac
volatile bool ITTic = false;              // Vrai si un Tic ou un Tac a été détecté (génération interruption). Doit être remis à faux par l'utilisateur
bool BAutoriseTicTacIT = false;           // Vrai si IT attaché à la pin signal

// Pin du processeur et usage
//
#define SignalEntree        9             // PIN Signal d'entrée de la mesure
#define LED_tic     4                     // PIN LED rouge (Tic)
#define LED_tac     5                     // PIN LED verte (Tac)

// Etats de l'automate
//
enum TEtatSynchro {E_Ne_Fait_Rien, E_Synchro_En_Cours,E_Synchro_En_Cours_Trans, E_Synchro_Ok, E_Synchro_Ok_Trans};
enum TEtatES {E_Attente,E_FenetreOccultationHaute,E_FenetreOccultationBasse};

/*
 *   Initialisation du programme
*/
void setup()
{

#ifdef DEBUG_SERIAL
  Serial.begin(9600);
#endif
  USART_Init_76800();               // Initialisation E/S séries avec carte affichage

  pinMode(SignalEntree, INPUT);     // La pin où est connecté le signal à mesurer
  pinMode(LED_tic, OUTPUT);         // La pin où est connecté la LED rouge
  pinMode(LED_tac, OUTPUT);         // La pin où est connecté la LED verte
  digitalWrite(LED_tic, HIGH);      // Tant que la fenêtre d'occultation n'est pas reçue, les 2 LEDS sont allumées et fixes.
  digitalWrite(LED_tac, HIGH);  
  attachPinChangeInterrupt(digitalPinToPinChangeInterrupt(SignalEntree), TicTacIT, RISING); // Signal d'entrée sous interruption mais non actif.
  BAutoriseTicTacIT = false;
  AutoriseTicTacIT(false);
  sei();
  
}

/*
   Procédure sous interruption
   Déclenchée si signal en entrée (tic-tac)
   ITTic (globale) est positionnée à vrai
   TempsTicTac = Temps courant au moment de l'interruption
*/

void TicTacIT()
{
  if (!ITTic) { // Si IT en cours de traitement, on l'ignore
    TempsTicTac = micros(); // Sauve la valeur du compteur courant
    ITTic = true;
  }
}

/*
 * Active et désactive les interruptions ITTic en fonction d'une variable d'entrée Autorise (vrai => active, false => desactive
 * La variable globale BAutoriseTicTacIT donne l'état courant (attaché ou pas) de l'interruption
 */
void AutoriseTicTacIT(bool Autorise) {
  if (!Autorise) {
    if (BAutoriseTicTacIT) {
      disablePinChangeInterrupt(digitalPinToPinChangeInterrupt(SignalEntree)); // On désactive l'interruption du signal d'entrée
      BAutoriseTicTacIT = false;
      ITTic = false;
    }
  } else {
    if (!BAutoriseTicTacIT) {
      enablePinChangeInterrupt(digitalPinToPinChangeInterrupt(SignalEntree)); // On active l'interruption du signal d'entrée
      BAutoriseTicTacIT = true;
      ITTic = false;
    }
  }
}

/*
 * Fait Basculer les LED en fonction d'un booléen en entrée
 * Ce booléen est inversé en sortie
*/
void FlipLed(bool *Etat)
{
  if (*Etat) { // Clignotement des LED
    digitalWrite(LED_tic, HIGH);
    digitalWrite(LED_tac, LOW);
  } else {
    digitalWrite(LED_tic, LOW);
    digitalWrite(LED_tac, HIGH);
  }
  *Etat  = !(*Etat);   // Pour affichage alterné des LED
}


//
// La fonction ci-dessous mesure le temps en prenant en compte un éventuel bouclage de la fonction millis ou micro
// On lui donne la valeur de départ de la valeur à mesurer et la valeur courante de millis ou micro. Elle rend la durée écoulée en millisecondes ou microseconde.
// DureeVrai = courante-dernière
//
unsigned long DureeVrai(unsigned long derniere, unsigned long courante) {
  if (courante < derniere) {  // on est passé par zéro
    return (0xffffffffu - derniere) + courante;
  } else {
    return courante - derniere;
  }
}


/*
 * Réception par test d'état ou via les routines standard selon
 * SERIAL_MANUEL. Pour la carte de mesure, il est impératif que les E/S
 * se fassent par test d'état => SERIAL_MANUEL doit être défini.
 * Pour la carte d'affichage, on peut utiliser l'un ou l'autre mode.
 * Utilisation de Serial1 (TX1 et RX1).
 */

 /*
  * Réception d'un caractère. La routine est boquante.
  */
 unsigned int USART_Receive()
{
#ifdef DEBUG_SERIAL
  unsigned int C;
  #ifdef SERIAL_MANUEL  
    while (!(UCSR1A & (1<<RXC1)));
    C = UDR1;
  #else
    while (Serial1.available() == 0);
    C = Serial1.read();
  #endif
    Serial.print("-> ");
    Serial.println(C,HEX);
    return C;  
#else
  #ifdef SERIAL_MANUEL  
    while (!(UCSR1A & (1<<RXC1)));
    return UDR1;
  #else
    while (Serial1.available() == 0);
    return Serial1.read();
  #endif
#endif
}

/*
 * Indique si un caractère a été reçu sur Serial1 (la fonction rend vrai dans ce cas, faux dans l'autre)
 */
bool USART_Available()
{
#ifdef SERIAL_MANUEL
  return (UCSR1A & (1<<RXC1));
#else
  return (Serial1.available() > 0);
#endif
}

/*
 * Attend jusqu'à ce que la transmission soit terminée (TXC1 = 1)
 */
void USART_Flush_Tx()
{
#ifdef SERIAL_MANUEL  
  while (!(UCSR1A & (1<<TXC1)));
  delay(5); // On laisse le temps à la ligne de se stabiliser
#else
  Serial1.flush();
#endif  
}

/*
 * Emission d'un caractère
 */
void USART_Transmit_Char(unsigned int data)
{
#ifdef SERIAL_MANUEL   
  while (!( UCSR1A & (1<<UDRE1))); // Tant que le buffer n'est pas vide, on boucle
  UDR1 = data;  // transmission
#else
  Serial1.write((byte)(data & 0xFF));
#endif  
#ifdef DEBUG_SERIAL
  Serial.print("<- ");
  Serial.print(((data >> 4) & 0x0F), HEX);
  Serial.println((data & 0x0F), HEX);
#endif
}
/**********************************
 * ROUTINES USART
**********************************/

// Init USART 
void USART_Init_76800()
{
#ifdef SERIAL_MANUEL  
  UCSR1B = 0;                         // Désactive UART en émission/réception
  UCSR1D = 0;                         // pas de CTS ni de RTS
  UCSR1C = (1<<UCSZ11)|(1<<UCSZ10);   // 8 bits 1 stop pas de parité
  //UCSR1A = (1<<U2X1);                 // double vitesse. Permet d'avoir une meilleure précision en bps
  UCSR1A = 0;
  UBRR1H = 0;                         // poids fort du diviseur
  //UBRR1L = 25;                        // = (16000000 / 4 / 76800 - 1 ) / 2
  UBRR1L = 12;                        // = (16000000 / 4 / 76800 - 1 ) / 4
  UCSR1B = (1<<RXEN1)|(1<<TXEN1);     // active l'émission et la réception
#else  
  Serial1.begin(76800,SERIAL_8N1);  // init E/S avec la carte pro micro qui réalise les mesures.
#endif  
}

/*
 * Emission Temps Tic Tac
 * En entrée : Temps Tic et Tac
 * La transmission se fait poids faible en tête pour chaque octet et poids faible en tête pour la trame
 * 150 => transmission 50 puis transmission 01
 * 39876 => transmission 76 puis transmission 98 puis transmission 03
 */
 void EmissionTempsTicTac(unsigned long T1,unsigned long T2) {
 unsigned int i;
#ifdef DEBUG_SERIAL
  Serial.print("Emet Temps : ");
  Serial.print(T1);
  Serial.print(" - ");
  Serial.println(T2);
#endif
  USART_Transmit_Char(ED1);
  do {
    i = (T1-((T1 / 10)*10)); // Attention aux optimisation !
    T1 = T1/10;
    i = ((T1-((T1 / 10)*10)) << 4) + i; // Attention aux optimisation !
    T1 = T1/10;
    USART_Transmit_Char(i);
  } while (T1 != 0);
  USART_Transmit_Char(ED2);
  do {
    i = (T2-((T2 / 10)*10)); // Attention aux optimisation !
    T2 = T2/10;
    i = ((T2-((T2 / 10)*10)) << 4) + i; // Attention aux optimisation !
    T2 = T2/10;
    USART_Transmit_Char(i);
  } while (T2 != 0);
  USART_Transmit_Char(EFinData);
 }


/*
 * APPLICATION MESURE DES TEMPS ET TRANSMISSION
 * Utilise un automate d'Etats finis
 * L'automate comporte des Etats et des transitions menant à l'Etat
 * convention de nommage : E_<nom d'état> représente un Etat
 *                         E_<nom d'état>_Trans représente une transition pour aller dans l'état <nom état>

 */

void loop()
{
  int i;
  unsigned int Recu;
  unsigned int V1; // pour calcul valeur fenetre occultation
  unsigned int Recu1; // pour calcul valeur fenetre occultation
  unsigned int Recu2; // pour calcul valeur fenetre occultation
  
  unsigned long FenetreOccultation;
  bool LEDFlip = true;   // Pour alterner l'allumage des LED
  unsigned long SynchroTimer;             // Timer pour la synchronisation et pour la détection d'inactivité
  unsigned long FenetreBasseOccultation = 0;  // Durée basse à partir de laquelle on accepte une interruption
  unsigned long FenetreHauteOccultation = 0;  // Durée haute au delà de laquelle on refuse accepte une interruption. On doit resynchroniser la mesure

  TEtatSynchro EtatSynchro = E_Ne_Fait_Rien; // Etat courant de l'automate de synchronisation
  TEtatES EtatES = E_Attente;             // Etat courant état automate réception

  while (1) {

    //-------------------------------------------------
    // Automate d'émission/réception
    //-------------------------------------------------
    /*
     * EN RECEPTION
     * ED1 <valeur Fenetre basse en milliseconde codé en DCB> ED2 <Valeur Fenetre haute en micro seconde codé en DCB>EFinAck
     * La routine met à jour la valeur de la fenêtre d'occultation.
     * Après un caractère EFinAck : la carte doit renvoyer EFinAck.
     *
     * EN EMISSION
     * ED1 <valeur Tic en micro seconde codé eb DCB> ED2 <Valeur Tac en micro seconde codé en DCB>EFinAck
     */



    switch (EtatES) {
       case E_Attente:
          if (USART_Available()) {
            Recu = USART_Receive();
            switch (Recu) {
              case ECartePresente:
                USART_Transmit_Char(ECartePresente);
              break;
              case ED1:
                EtatES = E_FenetreOccultationBasse;
                FenetreOccultation = 0;
                V1 = 1;
              break;
              default:
			        break;
            }
          }
		      break;  
		
       case  E_FenetreOccultationBasse: // Etat réception de la fenêtre d'occultation basse
          if (USART_Available()) {
            Recu = USART_Receive();
            switch (Recu) {
              case ED2:  // passage en etat Réception de la fenêtre d'occultation haute
                EtatES = E_FenetreOccultationHaute;
                FenetreBasseOccultation = FenetreOccultation;
                FenetreOccultation = 0;
                V1 = 1;
              break;
              case ECartePresente:
                USART_Transmit_Char(ECartePresente);
              break;
              default:
                Recu1 = Recu & 0x0F;
                Recu2 = (Recu >> 4) & 0x0F;
                if ((Recu1 <= 9) && (Recu2 <= 9)) {
                  FenetreOccultation = FenetreOccultation + (Recu1+Recu2*10)*V1;
                  V1 = V1*100;
                } else {
                  EtatES = E_Attente;
                }
              break;
            }
          }
       break;
	   
       case  E_FenetreOccultationHaute: // Etat réception de la fenêtre d'occultation basse
          if (USART_Available()) {
            Recu = USART_Receive();
            switch (Recu) {
               case EFinData: // Fin de réception des fenêtre d'occultation
                  FenetreHauteOccultation = FenetreOccultation;
                  if ((FenetreBasseOccultation > 0) && (FenetreHauteOccultation > 0)); { // on a un problème si une des valeur est nulle => On laisse tout tomber
  					        USART_Transmit_Char(EAckOccultation);
                    EtatSynchro = E_Synchro_En_Cours_Trans;   
#ifdef DEBUG_SERIAL
  Serial.print("Fenetre occultation :  ");
  Serial.print(FenetreBasseOccultation);
  Serial.print(" - ");
  Serial.println(FenetreHauteOccultation);
#endif                                  
  				        }
                  EtatES = E_Attente;
                break;
              case ECartePresente:
                USART_Transmit_Char(ECartePresente);
              break;
              default:
                Recu1 = Recu & 0x0F;
                Recu2 = (Recu >> 4) & 0x0F;
                if ((Recu1 <= 9) && (Recu2 <= 9)) {
                  FenetreOccultation = FenetreOccultation + (Recu1+Recu2*10)*V1;
                  V1 = V1*100;
                } else {
                  EtatES = E_Attente;
                }
              break;
            } 
          }
		   break;
       default:
         EtatES = E_Attente;
       break;
    }

    //-------------------------------------------------------------------------------------------
    // Automate de synchronisation. On ne le démarre que si on a reçu une fenêtre d'occultation
    //-------------------------------------------------------------------------------------------
    switch (EtatSynchro) {
      case E_Ne_Fait_Rien: // Pas de fenetre d'occultation. On ne fait rien
      break;
      
      case E_Synchro_En_Cours_Trans: // Passage en synchronisation. On recherche une période de silence suffisante pour que la prochaine IT soit le début d'un Tic ou d'un Tac
          SynchroTimer = millis();
          ITTic = false;
          EtatSynchro = E_Synchro_En_Cours;
          AutoriseTicTacIT(true); 
      break;

      case E_Synchro_En_Cours: //
        if (ITTic) { // Il y a eu une IT en dessous de la période attendue. Pas bon, on recommence à attendre un silence.
          FlipLed(&LEDFlip);
          SynchroTimer = millis();  // On réinitialise le timer
          ITTic = false;
        } else {
          if (DureeVrai(SynchroTimer,millis()) > DureeSansSignal) { //On n'a pas eu d'IT durant un certain temps.
            EtatSynchro = E_Synchro_Ok_Trans; // Donc, on a une période de silence suffisante pour que la prochaine IT soit le début d'un Tic ou d'un Tac.
          }
        }
      break;

      case E_Synchro_Ok_Trans: // On attend un début de Tic ou de Tac pour démarrer la mesure.
        if (ITTic) {
          AutoriseTicTacIT(false); // ***
          SynchroTimer = millis();                    // On réinitialise le timer
          EtatSynchro = E_Synchro_Ok;                 // Etat synchronisé
          TempsTicTacPrec = TempsTicTac;
          LEDFlip = false;
          ITTic = false;
        }
      break;


      case E_Synchro_Ok:
        if (DureeVrai(SynchroTimer,millis()) > FenetreHauteOccultation) {        // Si la différence entre le dernier Tic valide et le temps courant > à la fenêtre d'occultation haute, on a un pb. Il faut resynchroniser
            EtatSynchro = E_Synchro_En_Cours_Trans;
        } else {
          if (DureeVrai(SynchroTimer,millis()) > FenetreBasseOccultation) {     // Si la différence est supérieure à la fenetre basse d'occultation, on réactive les IT
            AutoriseTicTacIT(true); 
            if (ITTic) {
              AutoriseTicTacIT(false); 
              if (LEDFlip) {
                DiffTicTac = DureeVrai(TempsTicTacPrec,TempsTicTac);
                // Transmission de la valeur. La durée de la transmission est très faible par rapport à la durée de la mesure
                // Donc pas de risque de perdre des Tic ou des Tac pendant la transmission.
                EmissionTempsTicTac(DiffTicTacPrec,DiffTicTac);
              } else {
                DiffTicTacPrec = DureeVrai(TempsTicTacPrec,TempsTicTac);
              }
              TempsTicTacPrec = TempsTicTac;
              FlipLed(&LEDFlip);
              SynchroTimer = millis();
            }
          } else {
            AutoriseTicTacIT(false); // ***
          }
          ITTic = false;
        }
      break;
      default:
        EtatSynchro = E_Synchro_En_Cours_Trans;
      break;
    } // Switch
  } // While (1)
}
